package connect.demo;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.security.PublicKey;
import java.security.Security;
import java.security.Signature;
import java.util.Date;
import java.util.Properties;

import javax.servlet.ServletException;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.http.HttpResponse;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.impl.client.HttpClientBuilder;
import org.bouncycastle.jce.provider.BouncyCastleProvider;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;

/**
 * Servlet implementation class OpenidServlet
 * @author SAASPASS Developers.
 *
 */
@WebServlet("/SaaspassConnectServlet")
public class SaaspassConnectServlet extends HttpServlet {
	private static final long serialVersionUID = 1L;
	
	private static String redirectUrl;
	private static String saaspassServerUrl;
	private static String saaspassTokenEndPointPath;
	private static String saaspassUserinfoEndPointPath;
	private static String clientId;
	private static String clientSecret;
	private static String expectedIssuer;
	private static boolean validateIdTokenSignature;
	
	
    /**
     * @see HttpServlet#HttpServlet()
     */
    public SaaspassConnectServlet() {
        super();
        // TODO Auto-generated constructor stub
    }

    /**
	 * @see HttpServlet#doPost(HttpServletRequest request, HttpServletResponse response)
	 */
	protected void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
	}
	
	/**
	 * @see HttpServlet#doGet(HttpServletRequest request, HttpServletResponse response)
	 */
	protected void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
		
		/*
		 * Initialize security provider and load config parameters. Instead of doGet(), you can define an initialization point to load them based on your system needs.
		 */
		try {
			initializeConfiguration();
		} catch (SCException e2) {
			e2.printStackTrace();
			return;
		}
		
		/*
		 * First response contains authorization code and also state value that was generated on button click.
		 */
		String authorizationCode = request.getParameter("code");
		String stateValue = request.getParameter("state");
		
		boolean isValid = validateStateAndCode(request, stateValue, authorizationCode);
		if(!isValid) {
			System.out.println("code or state is invalid");
			//TODO ERROR: Here you should handle error and redirect user to error page or restart the flow.
			return;
		}
		else {
			System.out.println("Authorization code received.");
		}
		
		/*
		 * Get access token and id token.
		 */
		String accessToken = null;
		String idToken = null;
		try {
			JsonObject tokenResponseAsJson = getAccessTokenAndIdTokenResponse(authorizationCode);
			accessToken = tokenResponseAsJson.get("access_token").getAsString();		
			idToken = tokenResponseAsJson.get("id_token").getAsString();
			System.out.println("Access Token and Id Token received");
		} catch (URISyntaxException | SCException e1) {
			//TODO ERROR: token couldn't be received. Redirect user to an error page or restart flow to start with new state value generation. But investigate issue.
			System.out.println("Access Token couldn't be received.");
			e1.printStackTrace();
			return;
		}
		
		
		/*
		 * Validate Id Token signature.
		 * This is optional.
		 */
		if(validateIdTokenSignature) {
			boolean isValidSignature = validateIdTokenSignature(idToken);
			if(!isValidSignature) {
				System.out.println("id token signature not validated");
				//TODO ERROR: handle error case here. Redirect user to an error page. But investigate issue.
				return;
			}
			System.out.println("Id Token signature has been validated");
		}

		
		SaaspassUserProfile userProfile = null;
		
		/*
		 * Handle Id Token. 
		 * This is recommended.
		 */
		try {
			/*
			 * Id Token contains partial profile information. For full profile, you can call user info below.
			 */
			userProfile = handleIdToken(idToken);
			System.out.println("Id Token handled successfully");
		} catch (SCException e1) {
			// TODO ERROR: During handling id token. Handle error case here. Redirect user to an error page. But investigate issue.
			System.out.println("Handling Id Token failed.");
			e1.printStackTrace();
			return;
		}
		
		/*
		 * Get User Info. 
		 * This is optional. Call here if you need more user information than email.
		 */
		try {
			JsonObject userInfoData = getUserInfo(accessToken);
			updateUserProfile(userInfoData, userProfile);
			System.out.println("User authentication is successful and user info received.");
		} catch (URISyntaxException e) {
			//TODO ERROR: Cannot receive user info. Investigate issue here.
			e.printStackTrace();
			return;
		}
		
		if(userProfile != null) {
			System.out.println("------- USER INFO ---------");
			userProfile.printFields();
			//TODO now you can redirect user to your application landing page
			String userInfoAsParameters = "email="+userProfile.getEmail();
			response.setStatus(HttpServletResponse.SC_MOVED_TEMPORARILY);
			response.setHeader("Location", "applanding.html?"+userInfoAsParameters);
		}
	}
	
	/**
	 * Add BouncyCastle as a security provider and load Config parameters. You can load config in any other way you need.
	 * @throws SCException
	 */
	private void initializeConfiguration() throws SCException {
		
		Security.addProvider(new BouncyCastleProvider());
		System.out.println("security provider initialized");
		
		Properties configProperties;
		try {
			configProperties = Util.getConfigProperties();
		} catch (IOException e) {
			throw new SCException("couldn't load config parameters");
		}
		
		redirectUrl = configProperties.getProperty("your_redirect_url");
		clientId = configProperties.getProperty("your_client_id");
		clientSecret = configProperties.getProperty("your_client_secret");
		saaspassServerUrl = configProperties.getProperty("saaspass_url");
		saaspassTokenEndPointPath = configProperties.getProperty("saaspass_token_endpoint");
		saaspassUserinfoEndPointPath = configProperties.getProperty("saaspass_userinfo_endpoint");
		expectedIssuer = configProperties.getProperty("saaspass_issuer");
		validateIdTokenSignature = configProperties.getProperty("validate_idtoken_signature").equals("true");
		
		System.out.println("config parameters initialized");
	}
	
	/**
	 * Validate state value and authorization code.
	 * @param request
	 * @param state
	 * @param authorizationCode
	 * @return
	 */
	private boolean validateStateAndCode(HttpServletRequest request, final String state, final String authorizationCode) {
		String clientGeneratedState = (String)request.getSession().getAttribute("state");
		if(state == null || clientGeneratedState == null || authorizationCode == null) {
		    return false;
		}
		
		if(!state.equals(clientGeneratedState)) {
			return false;
		}
		
		return true;
	}
	
	/**
	 * Get Access Token and Id Token.
	 * @param authorizationCode
	 * @return
	 * @throws URISyntaxException
	 * @throws ClientProtocolException
	 * @throws IOException
	 * @throws SCException 
	 */
	public JsonObject getAccessTokenAndIdTokenResponse(String authorizationCode) throws URISyntaxException, ClientProtocolException, IOException, SCException {
		URI uri = new URIBuilder().setScheme("https")
				.setHost(saaspassServerUrl)
		        .setPath(saaspassTokenEndPointPath)
		        .setParameter("grant_type", "authorization_code")
		        .setParameter("code", authorizationCode)
		        .setParameter("redirect_uri", redirectUrl)
		        .setParameter("client_id", clientId)
		        .setParameter("client_secret", clientSecret)
		        .build();
		HttpPost postRequest = new HttpPost(uri);
		postRequest.setHeader("Content-type", "application/x-www-form-urlencoded");
		postRequest.setHeader("Accept", "application/json");
		
		HttpClient client = HttpClientBuilder.create().build();
		HttpResponse response = client.execute(postRequest);
		
		int responseCode = response.getStatusLine().getStatusCode();
		if(responseCode != 200) {
			throw new SCException("Error received from Saaspass");
		}
		
		BufferedReader rd = new BufferedReader(
                new InputStreamReader(response.getEntity().getContent()));

		StringBuffer result = new StringBuffer();
		String line = "";
		while ((line = rd.readLine()) != null) {
			result.append(line);
		}
	
		JsonParser parser = new JsonParser();
		JsonObject responseJson = (JsonObject)parser.parse(result.toString());
		
		return responseJson;
	}
	
	/**
	 * Handle Id Token.
	 * @param idToken
	 * @throws SCException 
	 */
	public SaaspassUserProfile handleIdToken(final String idToken) throws SCException {
		String[] idTokenParts = idToken.split("\\.");
	    
		//Headers
		/*
		 * Uncomment in case you want to log Jwt header:
		 */
		//String idTokenHeader = base64UrlDecode(idTokenParts[0]);
		//JsonParser parser = new JsonParser();	    
		//JsonObject headersJson = (JsonObject)parser.parse(idTokenHeader);
	    //String keyId = headersJson.get("kid").getAsString();
	    //String algorithm = headersJson.get("alg").getAsString();
	    
	    /*
	     * Claims as idToken payload data
	     */
		JsonParser parser = new JsonParser();
	    String idTokenDataAsJsonStr = Util.base64UrlDecode(idTokenParts[1]);
		JsonObject idTokenData = (JsonObject)parser.parse(idTokenDataAsJsonStr);
		
		//You may want to validate issuer
		String issuer = idTokenData.get("iss").getAsString();
		if(!issuer.equals(expectedIssuer)) {
			throw new SCException("unknown issuer received");
		}
		
		//You may want to validate client id
		String clientId = idTokenData.get("aud").getAsString();
		if(!clientId.equals(clientId)) {
			throw new SCException("invalid client id received");
		}
		
		/*
		 * You may want to validate Id token expiration time.
		 */
		Long expirationTime = idTokenData.get("exp").getAsLong();
		if(System.currentTimeMillis() < expirationTime) {
			throw new SCException("id token already expired");
		}
		
		/*
		 * Unique identifier of user's profile that was chosen by user during authentication. You may want to store this. 
		 * Note that SAASPASS users may have multiple profiles that can have different email or same email.
		 */
		String userProfileId = idTokenData.get("sub").getAsString();
		System.out.println("userProfileId: "+userProfileId);
		
		/*
		 * Email of authenticated user. It can be used as username of user for your application.
		 */
		String userEmail = idTokenData.get("email").getAsString();
		System.out.println("userEmail: "+userEmail);
		
		/*
		 * This is unique identifier of user's SAASPASS account. It is up to your need to use it in your system.
		 */
		String saaspassId = idTokenData.get("saaspass_id").getAsString();
		System.out.println("userSaaspassId: "+saaspassId);
		
		/*
		 * If you will call user info end point later, you don't use these attributes here yet.
		 */
		SaaspassUserProfile userProfile = new SaaspassUserProfile();
		userProfile.setProfileId(userProfileId);
		userProfile.setEmail(userEmail);
		userProfile.setSaaspassId(saaspassId);
		
		return userProfile;
	}
	
	/**
	 * Validate signature of Id Token. You need to have public key that is given on your SAASPASS Connect configuration.
	 * This is optional step. 
	 * @param idToken
	 * @return
	 */
	private boolean validateIdTokenSignature(final String idToken) {
		String[] idTokenParts = idToken.split("\\.");
		byte[] idTokenDataByteArr = (idTokenParts[0] + "." + idTokenParts[1]).getBytes(StandardCharsets.UTF_8);
		byte[] idTokenSignature = Util.base64UrlDecodeToBytes(idTokenParts[2]);
		
		try {
			PublicKey publicKey = Util.readPublicKey();
			boolean verified = verifyUsingPublicKey(idTokenDataByteArr, idTokenSignature, publicKey);
			return verified;
		} catch (GeneralSecurityException | IOException e) {
			e.printStackTrace();
			return false;
		}
	}
	
	/**
	 * Verification of signature via public key.
	 * @param data
	 * @param signature
	 * @param pubKey
	 * @return
	 * @throws GeneralSecurityException
	 */
    private static boolean verifyUsingPublicKey(byte[] data, byte[] signature, PublicKey pubKey) throws GeneralSecurityException {
        Signature sig = Signature.getInstance("SHA256withRSA");
        sig.initVerify(pubKey);
        sig.update(data);
        
        return sig.verify(signature);
    }

    /**
     * Get SAASPASS User info from userinfo end point that returns user's selected profile during login flow.
     * @param accessToken
     * @throws URISyntaxException
     * @throws ClientProtocolException
     * @throws IOException
     */
	public JsonObject getUserInfo(final String accessToken) throws URISyntaxException, ClientProtocolException, IOException {
		URI uri = new URIBuilder().setScheme("https")
				.setHost(saaspassServerUrl)
		        .setPath(saaspassUserinfoEndPointPath)
		        .build();
		HttpGet getRequest = new HttpGet(uri);
		getRequest.setHeader("Authorization", "Bearer "+accessToken);
		getRequest.setHeader("Accept", "application/jwt");
		
		HttpClient client = HttpClientBuilder.create().build();
		HttpResponse response = client.execute(getRequest);
		
		int responseCode = response.getStatusLine().getStatusCode();

		System.out.println("Response Code : " + responseCode);
		
		BufferedReader rd = new BufferedReader(
                new InputStreamReader(response.getEntity().getContent()));

		StringBuffer result = new StringBuffer();
		String line = "";
		while ((line = rd.readLine()) != null) {
			result.append(line);
		}
	
		
		String[] jwtResponseParts = result.toString().split("\\.");
		String userInfoAsJsonStr = Util.base64UrlDecode(jwtResponseParts[1]);
		JsonParser parser = new JsonParser();
		JsonObject userInfoData = (JsonObject)parser.parse(userInfoAsJsonStr);
		
		return userInfoData;
	}
	
	/**
	 * Setup user profile with the returned data from SAASPASS.
	 * @param userInfoData
	 */
	private void updateUserProfile(final JsonObject userInfoData, SaaspassUserProfile userProfile) {
		String profileId = userInfoData.get("sub").getAsString();
		userProfile.setProfileId(profileId);
		
		String email = userInfoData.get("email").getAsString();
		userProfile.setEmail(email);
		
		if(userInfoData.has("given_name")) {
			userProfile.setFirstName(userInfoData.get("given_name").getAsString());
		}
		
		if(userInfoData.has("family_name")) {
			userProfile.setLastName(userInfoData.get("given_name").getAsString());
		}
		
		if(userInfoData.has("name")) {
			userProfile.setFullName(userInfoData.get("name").getAsString());
		}
		
		if(userInfoData.has("gender")) {
			userProfile.setGender(userInfoData.get("gender").getAsString());
		}
		
		if(userInfoData.has("birthdate")) {
			userProfile.setBirthDate(userInfoData.get("birthdate").getAsString());
		}
		
		if(userInfoData.has("phone_number")) {
			userProfile.setPhoneNumber(userInfoData.get("phone_number").getAsString());
		}
		
		if(userInfoData.has("address")) {
			JsonObject addressInfoData = userInfoData.get("address").getAsJsonObject();
			if(addressInfoData.has("street_address")) {
				userProfile.setStreetAddress(addressInfoData.get("street_address").getAsString());
			}
			if(addressInfoData.has("locality")) {
				userProfile.setCity(addressInfoData.get("locality").getAsString());
			}
			if(addressInfoData.has("country")) {
				userProfile.setCountry(addressInfoData.get("country").getAsString());
			}
			if(addressInfoData.has("postal_code")) {
				userProfile.setPostalCode(addressInfoData.get("postal_code").getAsString());
			}
		}
		
		if(userInfoData.has("updated_at")) {
			userProfile.setUpdatedTime(new Date(userInfoData.get("updated_at").getAsLong()));
		}
		
		
	}
	
	

}
